from bson import ObjectId


class PyObjectId(ObjectId):
    '''
    Pydantic wrapper for bson ObjectIds
    '''
    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        if not ObjectId.is_valid(v):
            raise ValueError("Invalid objectid")
        return ObjectId(v)

    def __hash__(self):
        return hash(repr(self))



from typing import Dict, List, Optional, Union

from pydantic import BaseModel, validator

from ..media.text import split_size_str
from .base import PyObjectId
from .utils import POLYP_EVALUATION_STRUCTURE


class ReportPolypAnnotationResult(BaseModel):
    location_segment: Optional[str]
    location_cm: Optional[int]
    size_category: Optional[str]
    size_mm: Optional[int]
    surface_intact: Union[bool, str]
    rating: Optional[str]
    paris: Optional[List[str]]
    dysplasia: Optional[str]
    histo: Optional[str]
    morphology: Optional[str]
    nice: Optional[str]
    lst: Optional[str]
    non_lifting_sign: Union[bool, str] = "unknown"
    injection: Union[bool, str]

    tool: Optional[str]
    resection_technique: Optional[str]
    resection_status_microscopic: Optional[str]
    salvage: Optional[Union[bool, str]] = "unknown"

    apc_watts: Optional[int]
    number_clips: Optional[int]

    ectomy_wound_care_success: Optional[str]
    ectomy_wound_care_technique: Optional[str]
    ectomy_wound_care: Optional[Union[bool, str]]

    no_resection_reason: Optional[str]

    resection: bool

    polyp_id: Optional[PyObjectId]

    def __hash__(self):
        return hash(repr(self))

    @validator("paris", allow_reuse=True)
    def val_paris(cls, v):
        if v == []:
            v = None
        return v

    @validator("size_mm", allow_reuse=True)
    def val_size_mm(cls, v, values):
        if v:
            if v < 0:
                v = None
            else:
                _range = split_size_str(values["size_category"])
                assert _range
                assert _range[0] <= v <= _range[1]

        return v

    @validator("number_clips", "apc_watts", allow_reuse=True)
    def val_int(cls, v):
        if v:
            if v < 0:
                v = None
            return v

    @validator("resection", allow_reuse=True)
    def val_resection(cls, v, values):
        req_resection = [
            "resection_technique",
            "resection_status_microscopic",
            "salvage",
            "ectomy_wound_care",  # if true, ectomy_wound_care_success
        ]

        req_no_resection = ["no_resection_reason"]

        if v:
            for req in req_resection:
                assert req in values

        elif v is False:
            for req in req_no_resection:
                assert req in values

        return v

    @validator("ectomy_wound_care", allow_reuse=True)
    def val_ectomy_wound_care(cls, v, values):
        req_wound_care = ["ectomy_wound_care_success", "ectomy_wound_care_technique"]
        if v:
            for req in req_wound_care:
                assert req in values

        return v

    @validator("ectomy_wound_care_technique", allow_reuse=True)
    def val_ectomy_technique(cls, v, values):
        if v == "apc":
            assert "apc_watts" in values
        elif v == "clip":
            assert "number_clips" in values
        return v

    def evaluate(self):
        polyp_report = self.dict()
        result = {"required": [], "optional": [], "found": []}

        for attribute, element in POLYP_EVALUATION_STRUCTURE.items():
            if element["required"]:
                result["required"].append(attribute)

            else:
                required = False
                for requirement in element["required_if"]:
                    required_attribute = requirement["attribute"]
                    if polyp_report[required_attribute] in requirement["values"]:
                        required = True
                        break
                if required:
                    result["required"].append(attribute)
                else:
                    result["optional"].append(attribute)

        for key, value in polyp_report.items():
            if not value == None and not value == "unknown" and not value == -1:
                result["found"].append(key)

        result["required_found"] = [
            _ for _ in result["required"] if _ in result["found"]
        ]
        result["optional_found"] = [
            _ for _ in result["optional"] if _ in result["found"]
        ]

        result["required_missing"] = [
            _ for _ in result["required"] if _ not in result["found"]
        ]
        result["optional_missing"] = [
            _ for _ in result["optional"] if _ not in result["found"]
        ]

        return result


class ReportAnnotationResult(BaseModel):
    polyps: List[ReportPolypAnnotationResult]
    intervention_time: Optional[int]
    withdrawal_time: Optional[int]
    sedation: str
    bbps_worst: Optional[int]
    bbps_total: Optional[int]
    n_polyps: int
    n_adenoma: Optional[int]
    other_pathologies: bool
    indication: str
    mark_other: Optional[bool]

    def __hash__(self):
        return hash(repr(self))

    @validator("polyps", allow_reuse=True, pre=True)
    def dict_to_list(cls, v):
        if isinstance(v, dict):
            v = [__ for _, __ in v.items()]
        return v

    @validator(
        "intervention_time",
        "withdrawal_time",
        "bbps_worst",
        "bbps_total",
        "n_polyps",
        allow_reuse=True,
    )
    def del_unknown(cls, v):
        if v:
            if v < 0:
                v = None
        return v

    @validator("bbps_worst", allow_reuse=True)
    def bbps_worst_range(cls, v):
        if v:
            assert v <= 3
        return v

    @validator("bbps_total", allow_reuse=True)
    def bbps_total_range(cls, v):
        if v:
            assert v <= 9
        return v

    @validator("n_adenoma", allow_reuse=True)
    def set_n_adenoma(cls, v, values):
        return v


class Report:
    text: str
    structured: Optional[Dict[str, str]]



from datetime import datetime
from typing import List, Union

from pydantic import BaseModel, Field, validator

from ..classes.annotation import Flank


class ExternFlank(BaseModel):
    name: str = Field(alias="annotationName")
    value: Union[bool, str] = Field(alias="annotationValue")
    start: int = Field(alias="startFrame")
    stop: int = Field(alias="endFrame")

    @validator("name")
    def name_to_lower(cls, v):
        return v.lower()

    def to_intern(self, source, annotator_id, date, instance_id=None):
        return Flank(
            source=source,
            annotator_id=annotator_id,
            date=date,
            name=self.name,
            value=self.value,
            start=self.start,
            stop=self.stop,
            instance_id=instance_id,
        )


class ExternAnnotatedVideo(BaseModel):
    id_extern: int = Field(alias="videoId")
    video_key: str = Field(alias="videoName")
    date: datetime = Field(alias="lastAnnotationSession")

    @validator("date")
    def rm_tz(cls, v):
        v = v.replace(tzinfo=None)
        v = v.replace(microsecond=0)
        return v


class ExternVideoFlankAnnotation(BaseModel):
    extern_annotator_id: int = Field(alias="userId")
    id_extern: int = Field(alias="videoId")
    video_key: str = Field(alias="videoName")
    date: datetime = Field(alias="sessionDate")
    label_group_id: int = Field(alias="labelGroupId")
    label_group_name: str = Field(alias="labelGroupName")
    flanks: List[ExternFlank] = Field(alias="sequences")

    @validator("date")
    def rm_tz(cls, v):
        v = v.replace(microsecond=0)
        return v



from datetime import datetime
from typing import Dict, List, Optional, Union

from pydantic import BaseModel, Field, NonNegativeInt, validator

from .base import PyObjectId


class Annotation(BaseModel):
    source: str  # constr(regex=r"")
    annotator_id: NonNegativeInt
    date: datetime
    name: str
    instance_id: Optional[int]

    def __hash__(self):
        return hash(repr(self))

    @validator("date")
    def rm_ms(cls, v):
        v = v.replace(microsecond=0)
        return v


class MultilabelAnnotation(Annotation):
    value: List[NonNegativeInt]
    choices: List[str]
    instance_id: Optional[PyObjectId]


class ImageAnnotations(BaseModel):
    id: Optional[PyObjectId] = Field(alias="_id")
    image_id: Optional[PyObjectId]
    examination_id: Optional[PyObjectId]
    annotations: Dict[int, MultilabelAnnotation] = {}

    class Config:
        allow_population_by_field_name = True

    def latest_annotation(self):
        date = None
        for key, value in self.annotations.items():
            if not date:
                date = value.date
                selected_key = key
            if value.date > date:
                date = value.date
                selected_key = key

        return self.annotations[selected_key]

    def to_dict(self):
        _ = self.dict(exclude_none=True)
        _["annotations"] = {str(k): v for k, v in _["annotations"].items()}
        return _


class Flank(Annotation):
    start: int
    stop: int
    value: Union[bool, str]



import warnings
from pathlib import Path
from typing import Dict, List, Tuple
import numpy as np

from bson import ObjectId
from pymongo import MongoClient
from sklearn.model_selection import train_test_split

from .annotation import ImageAnnotations, MultilabelAnnotation
from .examination import Examination
from .image import Image, ImageCollection
from .model_handler import ModelHandler


def filter_label_array(array, target_labels, labels):
    """Label array of shape (n, m) where n is the number of images and m is the number of labels.\
        Selects columns of target labels and returns a new array of shape (n, len(target_labels))"""

    select = [labels.index(_) for _ in target_labels]
    array = array[:, select]
    return array


class DbHandler:
    def __init__(self, url):
        self.client = MongoClient(url, connectTimeoutMS=200, retryWrites=True)
        self.db = self.client.EndoData3
        self.examination = self.db.Examination
        self.image = self.db.Image
        self.image_collection = self.db.ImageCollection
        self.multilabel_annotation = self.db.MultilabelAnnotation
        self.video_segmentation = self.db.VideoSegmentation
        self.model = self.db.Model

    def clear_all(self):
        print("Deactivated")
        # self.examination.delete_many({})
        # self.image.delete_many({})
        # self.image_collection.delete_many({})

    def insert_frames(self, frames: List[Dict]):
        frame_dict = {}
        for frame in frames:
            r = self.image.insert_one(frame)
            frame_dict[str(frame["n"])] = r.inserted_id

        return frame_dict

    def insert_examination(self, examination, insert_frames=True):
        if examination.is_video:
            assert examination.path.exists()

        fps, frame_count = examination.get_video_info()
        examination.fps = fps
        examination.frame_count = frame_count
        r = self.examination.insert_one(examination.to_dict())
        assert r
        inserted_id = r.inserted_id
        examination.id = inserted_id

        frames = examination.generate_frame_list()
        frames = [Image(**_) for _ in frames]
        frames = [_.to_dict() for _ in frames]
        if insert_frames:
            frame_dict = self.insert_frames(frames)

        self.image_collection.insert_one(
            {"examination_id": examination.id, "type": "frame", "images": frame_dict}
        )

    def insert_model(self, model: ModelHandler):
        model_dict = model.to_dict(exclude_none=True)
        if "_id" in model_dict:
            r = self.model.update_one({"_id": model_dict["_id"]}, {"$set": model_dict})
        else:
            r = self.model.insert_one(model_dict)

        return r

    def update_examination_crop(self, _id: ObjectId, crop: Tuple[int]):
        self.examination.update_one({"_id": _id}, {"$set": {"crop": crop}})

    def update_frames_extracted(
        self, examination_id: ObjectId, frame_dict: Dict[int, Path]
    ):
        image_dict = self.image_collection.find_one(
            {"examination_id": examination_id, "type": "frame"}
        )
        for n, path in frame_dict.items():
            self.image.update_one(
                {"_id": image_dict["images"][str(n)]},
                {"$set": {"path": path.as_posix(), "is_extracted": True}},
            )

    def get_multilabel_train_data(self, test_size = 0.1):
        img_ids = [
            _["image_id"] for _ in self.multilabel_annotation.find({})
        ]
        train, val = train_test_split(img_ids, test_size=test_size)
        train = {i: _id for i, _id in enumerate(train)}
        val = {i: _id for i, _id in enumerate(val)}
        train_collection = ImageCollection(type="multilabel_train", images=train)
        train_collection_id = self.image_collection.insert_one(
            train_collection.to_dict()
        ).inserted_id

        val_collection = ImageCollection(type="multilabel_val", images=val)
        val_collection_id = self.image_collection.insert_one(
            val_collection.to_dict()
        ).inserted_id

        return train_collection_id, val_collection_id

    def get_examination(self, _id: ObjectId, as_object=True):
        examination = self.examination.find_one({"_id": _id})
        if as_object:
            examination = Examination(**examination)

        return examination

    def get_examinations(self, ids: List[ObjectId], as_object=True):
        examinations = self.examination.find({"_id": {"$in": ids}})
        if as_object:
            examinations = [Examination(**_) for _ in examinations]
        else:
            examinations = [_ for _ in examinations]

        return examinations

    def get_image(self, _id: ObjectId, as_object=True):
        image = self.image.find_one({"_id": _id})
        if as_object:
            image = Image(**image)

        return image

    def get_images(self, ids: List[ObjectId], as_object=True):
        images = self.image.find({"_id": {"$in": ids}})
        if as_object:
            images = [Image(**_) for _ in images]
        else:
            images = [_ for _ in images]

        return images

    def get_model_settings(self, model_id:ObjectId):
        settings = self.model.find_one({"_id": model_id})
        trainer_settings = settings["trainer"]
        model_settings = settings["model"]

        return model_settings, trainer_settings

    def prepare_ds_from_image_collection(self, image_collection_id, predict=False, target_labels=None):
        image_collection = self.image_collection.find_one({"_id": image_collection_id})
        paths = []
        _ids = []
        labels = []
        crop = []
        choices = []

        if image_collection:
            images = self.image.find(
                {
                    "_id": {"$in": list(image_collection["images"].values())},
                    "is_extracted": True,
                }
            )
            # ids = [_["_id"] for _ in images]
            examination_ids = self.image.distinct(
                "examination_id",
                {"_id": {"$in": list(image_collection["images"].values())}},
            )
            lookup_crop = {}
            for examination_id in examination_ids:
                _ = self.examination.find_one({"_id": examination_id})
                if "crop" in _:
                    lookup_crop[examination_id] = _["crop"]
                else:
                    warnings.warn(f"Examination {examination_id} has no crop")

            for image in images:
                if image["examination_id"] in lookup_crop:
                    crop.append(lookup_crop[image["examination_id"]])
                    paths.append(image["path"])
                    _ids.append(image["_id"])
                else:
                    continue

        if predict:
            labels = [str(_) for _ in _ids]
        else:
            assert target_labels

            for _id in _ids:
                annotation = self.get_multilabel_image_annotation(_id)
                annotation = annotation.latest_annotation()
                if not choices:
                    choices = annotation.choices
                else:
                    assert choices == annotation.choices

                labels.append(annotation.value)

            label_records = []
            for label in labels:
                record = []
                for i, _ in enumerate(choices):
                    if i in label:
                        record.append(1.0)
                    else: record.append(0.0)
                label_records.append(record)

            labels = np.array(label_records, dtype = float)
            labels = filter_label_array(labels, target_labels, choices)

        return paths, labels, crop, choices

    def get_examination_image_collection(self, examination_id: ObjectId):
        image_collection = self.image_collection.find_one(
            {"examination_id": examination_id}
        )
        return ImageCollection(**image_collection)

    def get_examination_frame_dict(self, examination_id: ObjectId):
        image_collection = self.image_collection.find_one(
            {"examination_id": examination_id, "type": "frame"}
        )
        if image_collection:
            return image_collection["images"]

        return {}

    def get_multilabel_image_annotation(self, image_id: ObjectId):
        image = self.multilabel_annotation.find_one({"image_id": image_id})
        if image:
            return ImageAnnotations(**image)

        return {}

    def get_multilabel_image_annotations(self, image_ids: List[ObjectId]):
        images = self.multilabel_annotation.find({"image_id": {"$in": image_ids}})
        images = [ImageAnnotations(**image) for image in images]

        return images

    def set_image_annotation(
        self, image_id: ObjectId, annotation: MultilabelAnnotation
    ):
        r = self.multilabel_annotation.update_one(
            {"_id": image_id},
            {"$set": {str(annotation.annotator_id): annotation.to_dict()}},
        )
        assert r.matched_count == 1
        return r

    # def get_report(self, _id: ObjectId, as_object = True):
    #     report = self.ASDASDASDASDASD.find_one({"examination_id": _id, "type": "report"})
    #     if as_object:
    #         report = ImageCollection(**report)

    #     return report



from sklearn.metrics import precision_score, recall_score, f1_score
import numpy as np

# def get_classification_report(results_df, prediction_type):
#     s1 = results_df._type == "annotations"
#     y_true = results_df[s1].value.to_list()
#     s1 = results_df._type == prediction_type
#     y_pred = results_df[s1].value.to_list()
#     _report = classification_report(y_true, y_pred, output_dict=True)
#     return _report


def calculate_metrics(pred, target, threshold=0.5):
    pred = np.array(pred > threshold, dtype=float)
    return {'micro/precision': precision_score(y_true=target, y_pred=pred, average='micro', zero_division = 0),
            'micro/recall': recall_score(y_true=target, y_pred=pred, average='micro', zero_division = 0),
            'micro/f1': f1_score(y_true=target, y_pred=pred, average='micro', zero_division = 0),
            'macro/precision': precision_score(y_true=target, y_pred=pred, average='macro', zero_division = 0),
            'macro/recall': recall_score(y_true=target, y_pred=pred, average='macro', zero_division = 0),
            'macro/f1': f1_score(y_true=target, y_pred=pred, average='macro', zero_division = 0),
            'samples/precision': precision_score(y_true=target, y_pred=pred, average=None, zero_division = 0),
            'samples/recall': recall_score(y_true=target, y_pred=pred, average=None, zero_division = 0),
            'samples/f1': f1_score(y_true=target, y_pred=pred, average=None, zero_division = 0),
            }


import requests
from urllib3.exceptions import InsecureRequestWarning
from .classes import ExternAnnotatedVideo, ExternVideoFlankAnnotation

requests.packages.urllib3.disable_warnings(category=InsecureRequestWarning)

def get_extern_annotations(url, auth):
    r = requests.get(f"{url}/GetVideosWithAnnotations", auth=auth, verify=False)
    assert r.status_code == 200
    r = [ExternAnnotatedVideo(**_) for _ in r.json()]

    return r

def get_extern_video_annotation(video_key, url, auth):
    r = requests.get(url+"/GetAnnotationsByVideoName/"+video_key, auth=auth, verify=False)
    assert r.status_code == 200
    annotations = [ExternVideoFlankAnnotation(**_) for _ in r.json()]

    return annotations

def get_smartie_data(url, auth):
    r = requests.get(url+"/GetSmartieVideoData", auth = auth, verify=False)
    assert r.status_code == 200

    r = r.json()
    return r



POLYP_EVALUATION_STRUCTURE = {
        "location_segment": {
        "attribute": "location_segment",
        "required": True,
        "required_if": []
        },
        "size_category": {
        "attribute": "size_category",
        "required": True,
        "required_if": []
        },
        "surface_intact": {
        "attribute": "surface_intact",
        "required": True,
        "required_if": [],
        },
        "rating": {
        "attribute": "rating",
        "required": True,
        "required_if": [],
        },
        "resection": {
        "attribute": "resection",
        "required": True,
        "required_if": [],
        },
        "location_cm": {
        "attribute": "location_cm",
        "required": False,
        "required_if": [
            {
            "attribute": "location_segment",
            "values": [
                "rectum",
                "sigma"
            ]
            }
        ],
        },
        "size_mm": {
        "attribute": "polyp_size_mm",
        "required": False,
        "required_if": []
        },
        "paris": {
        "attribute": "polyp_paris",
        "required": False,
        "required_if": [
            {
            "attribute": "size_category",
            "values": [
                "<5",
                "5-10",
                ">10-20",
                ">20"
            ]
            }
        ]
        },
        "lst": {
        "attribute": "lst",
        "required": False,
        "required_if": [
            {
            "attribute": "size_category",
            "values": [
                ">10-20",
                ">20"
            ]
            }
        ],
        },
        "nice": {
        "attribute": "nice",
        "required": False,
        "required_if": [
            {
            "attribute": "size_category",
            "values": [
                "5-10",
                ">10-20",
                ">20"
            ]
            }
        ],
        },
        "non_lifting_sign": {
        "attribute": "non_lifting_sign",
        "required": False,
        "required_if": [],
        },
        "tool": {
        "attribute": "tool",
        "required": False,
        "required_if": [
            {
            "attribute": "resection",
            "values": [
                True
            ]
            }
        ],
        },
        # "resection_technique": {
        # "attribute": "resection_technique",
        # "required": False,
        # "required_if": [
        #     {
        #     "attribute": "resection",
        #     "values": [
        #         True
        #     ]
        #     }
        # ],
        # },
        "salvage": {
        "attribute": "salvage",
        "required": False,
        "required_if": [
            {
            "attribute": "resection",
            "values": [
                True
            ]
            }
        ],
        },
        "ectomy_wound_care": {
        "attribute": "ectomy_wound_care",
        "required": False,
        "required_if": [
            {
            "attribute": "resection",
            "values": [
                True
            ]
            }
        ],
        },
        "ectomy_wound_care_technique": {
        "attribute": "ectomy_wound_care_technique",
        "required": False,
        "required_if": [
            {
            "attribute": "ectomy_wound_care",
            "values": [
                True
            ]
            }
        ],
        },
        "apc_watts": {
        "attribute": "apc_watts",
        "required": False,
        "required_if": [
            {
            "attribute": "ectomy_wound_care_technique",
            "values": [
                "apc"
            ]
            }
        ],
        },
        "number_clips": {
        "attribute": "number_clips",
        "required": False,
        "required_if": [
            {
            "attribute": "ectomy_wound_care_technique",
            "values": [
                "clip"
            ]
            }
        ],
        },
        "ectomy_wound_care_success": {
        "attribute": "ectomy_wound_care_success",
        "required": False,
        "required_if": [
            {
            "attribute": "ectomy_wound_care",
            "values": [
                True
            ]
            }
        ],
        },
        "no_resection_reason": {
        "attribute": "no_resection_reason",
        "required": False,
        "required_if": [
            {
            "attribute": "resection",
            "values": [
                False
            ]
            }
        ],
        },
        "resection_technique": {
        "attribute": "resection_technique",
        "required": False,
        "required_if": [
            {
            "attribute": "resection",
            "values": [
                True
            ]
            }
        ]
        }
        }
        


import numpy as np
import wandb
from pytorch_lightning import Callback, Trainer
from pytorch_lightning.loggers import LoggerCollection, WandbLogger
from torch import sigmoid


def get_wandb_logger(trainer: Trainer) -> WandbLogger:
    """Safely get Weights&Biases logger from Trainer."""

    if isinstance(trainer.logger, WandbLogger):
        return trainer.logger

    if isinstance(trainer.logger, LoggerCollection):
        for logger in trainer.logger:
            if isinstance(logger, WandbLogger):
                return logger

    raise Exception(
        "You are using wandb related callback, but WandbLogger was not found for some reason..."
    )


class WatchModelWithWandb(Callback):
    """Make WandbLogger watch model at the beginning of the run."""

    def __init__(self, log: str = "gradients", log_freq: int = 100):
        self.log = log
        self.log_freq = log_freq

    def on_train_start(self, trainer, pl_module):
        logger = get_wandb_logger(trainer=trainer)
        logger.watch(model=trainer.model, log=self.log, log_freq=self.log_freq)


class ImagePredictionLogger(Callback):
    """Logs a validation batch and their predictions to wandb.
    Example adapted from:
        https://wandb.ai/wandb/wandb-lightning/reports/Image-Classification-using-PyTorch-Lightning--VmlldzoyODk1NzY
    """

    def __init__(self, num_samples: int = 8):
        super().__init__()
        self.num_samples = num_samples
        self.ready = True

    def on_sanity_check_start(self, trainer, pl_module):
        self.ready = False

    def on_sanity_check_end(self, trainer, pl_module):
        """Start executing this callback only after all validation sanity checks end."""
        self.ready = True

    def on_validation_epoch_end(self, trainer, pl_module):
        if self.ready:
            logger = get_wandb_logger(trainer=trainer)
            experiment = logger.experiment

            # print("_________________________-")
            # print(vars(trainer))
            # get a validation batch from the validation dat loader
            val_samples = next(iter(trainer.val_dataloaders[0]))
            val_imgs, _val_labels = val_samples

            # run the batch through the network
            val_imgs = val_imgs.to(device=pl_module.device)
            logits = pl_module(val_imgs).to("cpu")
            # preds = torch.argmax(logits, axis=-1)
            preds = sigmoid(logits).numpy()  # .squeeze()

            _label_list = pl_module.labels

            preds[preds >= 0.5] = 1
            preds[preds < 0.5] = 0
            preds = np.argwhere(preds)

            pred_labels = {i: [] for i in range(len(val_imgs))}
            for _pred in preds:
                pred_labels[_pred[0]].append(_label_list[_pred[1]])
            pred_labels = [pred_labels[i] for i in range(len(val_imgs))]

            _val_labels = _val_labels.to("cpu").numpy()
            _val_labels = np.argwhere(_val_labels)
            val_labels = {i: [] for i in range(len(val_imgs))}
            for _label in _val_labels:
                val_labels[_label[0]].append(_label_list[_label[1]])

            val_labels = [val_labels[i] for i in range(len(val_imgs))]

            # log the images as wandb Image
            experiment.log(
                {
                    f"Images/{experiment.name}": [
                        wandb.Image(x, caption=f"Pred:{pred}, Label:{y}")
                        for x, pred, y in zip(
                            val_imgs[: self.num_samples],
                            pred_labels[: self.num_samples],
                            val_labels[: self.num_samples],
                        )
                    ]
                }
            )



from datetime import datetime
from pathlib import Path
from typing import Dict, Optional

from pydantic import BaseModel, Field

from .base import PyObjectId


class Image(BaseModel):
    id: Optional[PyObjectId] = Field(alias="_id")
    examination_id: PyObjectId
    origin: str
    origin_category: str
    image_type: str # freeze, frame
    n: int
    is_extracted: bool
    annotation: Optional[PyObjectId]
    prediction: Optional[PyObjectId]
    path: Optional[Path]

    class Config:
        allow_population_by_field_name = True
        # arbitrary_types_allowed = True
        json_encoders = {Path: str}

    def to_dict(self):
        _ = self.dict(exclude_none=True)
        if self.path:
            _["path"] = str(self.path)
        return _

    def exists(self):
        if self.is_extracted:
            return self.path.exists()
        else:
            return False

class ImageCollection(BaseModel):
    id: Optional[PyObjectId] = Field(alias="_id")
    examination_id: Optional[PyObjectId]
    model_id: Optional[PyObjectId]
    type: Optional[str]
    images: Dict[int, PyObjectId]
    date: Optional[datetime] = datetime.now()

    def to_dict(self):
        _ = self.dict(exclude_none=True)
        _["images"] = {str(k):v for k,v in _["images"].items()}
        return _



from typing import (
    List,
    Optional,
    Tuple
)
from pathlib import Path
from pydantic import (
    BaseModel,
    Field,
)

from .base import PyObjectId
from ..media.video import get_video_info

class Examination(BaseModel):
    id: Optional[PyObjectId] = Field(alias="_id")
    origin: str
    origin_category: str
    examination_type: str
    is_video: bool
    video_key: Optional[str]
    id_extern: Optional[int]
    examiners: Optional[List[str]]
    age: Optional[int]
    gender: Optional[int]
    path: Optional[Path]
    fps: Optional[int]
    frame_count: Optional[int]
    frames: Optional[PyObjectId]
    freezes: Optional[PyObjectId]
    report: Optional[PyObjectId]
    segmentation: Optional[PyObjectId]
    annotation: Optional[PyObjectId]
    prediction: Optional[PyObjectId]
    crop: Optional[Tuple[int, int, int, int]] # ymin, ymax, xmin, xmax

    class Config:
        allow_population_by_field_name = True
        # arbitrary_types_allowed = True
        json_encoders = {Path: str}

    def to_dict(self):
        _dict = self.dict(exclude_none=True)
        _dict["path"] = str(self.path)

        return _dict

    def get_video_info(self):
        """
        Get video info from a video file.

        Returns fps, frame_count
        """
        return get_video_info(self.path)

    def get_frame_template(self):
        template = {
            "examination_id": self.id,
            "origin": self.origin,
            "origin_category": self.origin_category,
            "image_type": "frame",
            "is_extracted": False
        }

        return template

    def generate_frame_list(self):
        """
        Generate a list of frames from a video.
        """
        template = self.get_frame_template()
        frames = []
        self.fps, self.frame_count = self.get_video_info()
        for i in range(self.frame_count):
            _ = template.copy()
            _["n"] = i
            frames.append(_)

        return frames


from datetime import datetime as dt
from pathlib import Path

from faker import Faker

from .annotation import MultilabelAnnotation
from .examination import Examination
from .report import ReportAnnotationResult, ReportPolypAnnotationResult


class DataGen:
    def __init__(self, data_dir: Path):
        self.factory = Faker()
        self.dir = data_dir

    def text(self):
        return self.factory.text()

    def examination_with_video(self):
        examination = Examination(
            origin = "test",
            origin_category = "test",
            is_video = True,
            path = self.dir.joinpath('test_video.mp4'),
            examination_type = "unknown",
        )
        assert examination.path.exists()
        return examination

    def multilabel_annotation(self):
        annotation = {
            "source": "test",
            "annotator_id": 999,
            "date": dt.now(),
            "name": "test",
            "value": 1,
            "choices": ["test1", "test2", "test3"],
        }

        return MultilabelAnnotation(**annotation)

    def polyp_report_annotation(self):
        report = {
            "location_segment": "sigma",
            "location_cm": 16,
            "size_category": "<5mm",
            "resection": False
        }

        report = ReportPolypAnnotationResult(**report)
        return report

    def report_annotation(self):
        report = {
            "polyps": [self.polyp_report_annotation()],
            "withdrawal_time": 300, # in s
            "n_polyps": 1,
            "indication": "screening"
        }

        return ReportAnnotationResult(**report)







from bson import ObjectId
from ..classes.video_segmentation import VideoSegmentation

def convert_extern_annotation(annotation):
    # video_segmentation = {}
    annotation_by_label = {}
    label_annotation_index = {}
    dates = [_.date for _ in annotation]
    max_date = max(dates)

    # create a dict with all labels and the corresponding annotation index
    for i, _annotation in enumerate(annotation):
        for value in _annotation.flanks:
            if not value.name in label_annotation_index:
                label_annotation_index[value.name] = []
            label_annotation_index[value.name].append(i)

    # select the most recent annotation for each label
    for key in label_annotation_index.keys():
        label_annotation_index[key] = list(set(label_annotation_index[key]))
        indices = label_annotation_index[key]
        selected_dates = [dates[i] for i in indices]
        max_date_index = selected_dates.index(max(selected_dates))
        selected_index = indices[max_date_index]
        selected_annotation = annotation[selected_index]
        annotation_by_label[key] = selected_annotation

    # filter all flanks, so that only flanks of the corresponding label are in the dict entry
    flanks = []
    for key in annotation_by_label.keys():
        _annotation = annotation_by_label[key]
        annotator_id = _annotation.extern_annotator_id
        date = _annotation.date
        values = _annotation.flanks
        values = [_ for _ in values if _.name == key]
        values = [_.to_intern(
            source = "video_web_annotation",
            annotator_id = annotator_id,
            date = date
        ) for _ in values]
        flanks.extend(values)

    flanks.sort(key=lambda x: x.start)
    return flanks, max_date

def extern_to_intern_video_annotation(examination_id: ObjectId, annotation):
    flanks, max_date = convert_extern_annotation(annotation)
    segmentation_dict = {
        "examination_id": examination_id,
        "annotation": flanks
    }
    segmentation = VideoSegmentation(**segmentation_dict)

    return segmentation


from pydantic import BaseModel, Field
from typing import Dict, List, Optional, Union
from .base import PyObjectId

class Prediction(BaseModel):
    model_id: PyObjectId
    image_id: PyObjectId
    prediction: List[float]
    choices: List[str]
    labels: List[List[str]]

    


from typing import List, Optional

from pydantic import BaseModel, Field

from .annotation import Flank
from .base import PyObjectId


class VideoSegmentation(BaseModel):
    id: Optional[PyObjectId] = Field(alias="_id")
    examination_id: Optional[PyObjectId]
    annotation: List[Flank] = []
    prediction_raw: List[Flank] = []
    prediction_smooth: List[Flank] = []
    prediction: List[Flank] = []

    def to_dict(self):
        _ = self.dict(exclude_none=True, exclude_defaults=True)
        return _

    class Config:
        allow_population_by_field_name = True



from pathlib import Path

from pydantic import BaseModel
from tqdm import tqdm

from .db import DbHandler


class DbValidator(BaseModel):
    db: DbHandler

    def get_missing_extracted_imgs(self):
        imgs = self.db.image.find({"is_extracted": True})
        i = 0
        failed = []
        for img in tqdm(imgs):
            if i == 10000:
                i = 0
                _ = self.db.image.find_one({})  # keep Db Alive
            i += 1
            path = Path(img["path"])
            if not path.exists():
                failed.append(img)

        return failed



from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel, Field

from ..model.multilabel_classification_net import MultilabelClassificationNet
from .base import PyObjectId
from .image import ImageCollection

MODEL_LOOKUP = {
    "colo_segmentation": MultilabelClassificationNet
}

class ModelHandler(BaseModel):
    id: Optional[PyObjectId] = Field(alias="_id")
    date: datetime = datetime.now()
    path: Optional[Path]
    model_name: str # e.g. colo_segmentation
    model_type: str # e.g. EfficientNet-B4
    labels: Optional[List[str]]
    train_images: Optional[ImageCollection] # change to id
    val_images: Optional[ImageCollection] # change to id
    model_settings: Optional[Dict[str, Any]]
    metrics: Any

    """
    Example model_settings:
    {
        labels = ["background", "foreground"],
        lr=6e-3,
        wigth_decay=1e-5,
        pos_weight=2,
        model_type="RegNetX800MF"
    }
    """

    class Config:
        allow_population_by_field_name = True

    def load_model_from_checkpoint(self, model_path, gpu = True, eval = True):

        model_class = MODEL_LOOKUP[self.model_name]
        model = model_class.load_from_checkpoint(model_path)
        if gpu:
            model.cuda()
        if eval:
            model.eval()

        return model

    def prepare_model(self, gpu = True, eval = False):
        assert hasattr(self, "model_settings"), "ModelHandler has no model_settings attribute"

        model_class = MODEL_LOOKUP[self.model_name]
        settings = self.model_settings.copy()
        settings["labels"] = self.labels
        settings["model_type"] = self.model_type
        model = model_class(**settings)
        if gpu:
            model.cuda()
        if eval:
            model.eval()

        return model, settings



# train_ds, val_ds, train_collection_id, val_collection_id = db.get_multilabel_train_data()
from pathlib import Path
from typing import Any, Dict, List, Optional

import pytorch_lightning
from pydantic import BaseModel
from torch.utils.data import DataLoader
from ..model import wandb_callbacks
from ..model.utils import instantiate

from ..dataset.image_classification import ImageClassificationDs
from .base import PyObjectId
from .db import DbHandler


class TrainerHandler(BaseModel):
    model_name: str = "colo_segmentation"
    model_type: str = "RegNetX800MF"

    target_labels: List[str]

    train_collection_id: PyObjectId
    val_collection_id: PyObjectId

    train_bs: int
    val_bs: int

    train_workers: int
    val_workers: int

    scaling: int

    checkpoint_dir: Path

    gpus: int = 1

    log_dir: Path
    log_freq: int = 100

    save_top_k: int = 1
    monitor: str = "val/loss"
    mode: str = "min"
    early_stopping_patience: int = 25
    max_epochs: int = 300

    trainer_precision:int = 16

    train_ds: Any = None
    val_ds: Any = None
    train_dl: Any = None
    val_dl: Any = None
    model_path: Optional[Path] = None

    callbacks: Dict[str, Any] = {
        "model_checkpoint": {
            "_target_": pytorch_lightning.callbacks.ModelCheckpoint,
            "monitor": None,
            "save_top_k": None,
            "save_last": False,
            "mode": None,
            "dirpath": None,
            "filename": None,
        },
        "early_stopping": {
            "_target_": pytorch_lightning.callbacks.EarlyStopping,
            "monitor": None,
            "patience": None,
            "mode": None,
        },
        "watch_model_with_wandb": {
            "_target_": wandb_callbacks.WatchModelWithWandb,
            "log": "all",
            "log_freq": None,
        },
        "image_prediction_logger": {
            "_target_": wandb_callbacks.ImagePredictionLogger,
            "num_samples": 10
        }
    }

    def get_dataloaders(self, db:DbHandler):
        train_dl = self.prepare_train_data(db)
        val_dl = self.prepare_val_data(db)

        return train_dl, val_dl

    def get_settings(self):
        settings = self.dict(exclude_none = True, exclude= {
            "train_ds",
            "val_ds",
            "train_dl",
            "val_dl",
            "model_filename",
            "callbacks"
        })
        if "model_path" in settings:
            settings["model_path"] = self.model_path.as_posix()
        if "log_dir" in settings:
            settings["log_dir"] = self.log_dir.as_posix()
        if "checkpoint_dir" in settings:
            settings["checkpoint_dir"] = self.checkpoint_dir.as_posix()
        return settings

    def prepare_trainer(self):
        logger, callbacks = self.instantiate_callbacks()
        trainer = pytorch_lightning.Trainer(
            max_epochs=self.max_epochs,
            gpus=self.gpus,
            logger=logger,
            callbacks=callbacks,
            precision=self.trainer_precision,
        )
        settings = self.get_settings()
        return trainer, settings

    def get_model_filename(self):
        i = 0
        filename = f"{self.model_name}_{self.model_type}_{i}.ckpt"
        path = self.checkpoint_dir.joinpath(filename)
        exists = path.exists()
        while exists:
            i += 1
            filename = f"{self.model_name}_{self.model_type}_{i}.ckpt"
            path = self.checkpoint_dir.joinpath(filename)
            exists = path.exists()
                
        self.model_path = path
        return filename

    def dump_config(self):
        pass

    def prepare_callbacks(self):
        self.get_model_filename()
        
        callbacks = self.callbacks.copy()
        callbacks["model_checkpoint"]["dirpath"] = self.checkpoint_dir
        callbacks["model_checkpoint"]["filename"] = self.model_path.with_suffix("").name
        callbacks["model_checkpoint"]["monitor"] = self.monitor
        callbacks["model_checkpoint"]["save_top_k"] = self.save_top_k
        callbacks["model_checkpoint"]["mode"] = self.mode

        callbacks["early_stopping"]["monitor"] = self.monitor
        callbacks["early_stopping"]["patience"] = self.early_stopping_patience
        callbacks["early_stopping"]["mode"] = self.mode

        callbacks["watch_model_with_wandb"]["log_freq"] = self.log_freq

        self.callbacks = callbacks
        return callbacks

    def instantiate_callbacks(self, include_wandb=True):
        if include_wandb:
            logger = {
                "_target_": pytorch_lightning.loggers.wandb.WandbLogger,
                "project": self.model_name,
                "job_type": "train",
                "group": "",
                "save_dir": self.log_dir,
            }
            logger = instantiate(logger)
        else: logger = None

        callbacks = self.prepare_callbacks()
        skip_callbacks = []
        if not include_wandb: 
            skip_callbacks.append("watch_model_with_wandb")
            skip_callbacks.append("image_prediction_logger")

        callbacks = [instantiate(v) for k, v in callbacks.items()]
        return logger, callbacks


    def prepare_train_data(self, db):
        paths, labels, crop, choices = db.prepare_ds_from_image_collection(
            self.train_collection_id,
            predict = False,
            target_labels = self.target_labels,
            )
        self.train_ds = ImageClassificationDs(
            paths = paths, 
            labels = labels,
            crop = crop,
            scaling = self.scaling,
            training = True,
        )
        self.train_dl = DataLoader(self.train_ds, batch_size = self.train_bs, num_workers = self.train_workers, shuffle = True)
        return self.train_dl

    def prepare_val_data(self, db):
        paths, labels, crop, choices = db.prepare_ds_from_image_collection(
            self.val_collection_id,
            predict = False,
            target_labels = self.target_labels,
            )
        self.val_ds = ImageClassificationDs(
            paths = paths, 
            labels = labels,
            crop = crop,
            scaling = self.scaling,
            training = True,
        )
        self.val_dl = DataLoader(self.val_ds, batch_size = self.val_bs, num_workers = self.val_workers, shuffle = False)
        return self.val_dl
